@@ -11,7 +11,7 @@ __all__ = [
 ]
 
 
-def get_app(dispatcher):
+def get_app(dispatcher, tag):
     """
     Return the manage app from the dispatcher.
 
@@ -20,7 +20,7 @@ def get_app(dispatcher):
     cases.
     """
     for app in dispatcher.apps:
-        if hasattr(app, 'blueprints') and '_manhattan_manage' in app.blueprints:
+        if hasattr(app, '_manage') and app._manage._tag == tag:
             return app
 
 def get_config(app, blueprint_name):
