@@ -6,21 +6,24 @@ Link factories for manhattan views.
 import bson
 import json
 import os
-import urllib
+import urllib.parse as urlparse
+from urllib.parse import urlencode
 
 import flask
 import inflection
 from manhattan import secure
-from manhattan.assets.fields import AssetField
-from manhattan.assets.transforms.base import BaseTransform
 from manhattan.nav import Nav, NavItem
 from manhattan.manage.views import utils
 
 # Optional imports
 try:
     from manhattan.assets import Asset
+    from manhattan.assets.fields import AssetField
+    from manhattan.assets.transforms.base import BaseTransform
+    __assets_supported__ = True
+
 except ImportError as e:
-    pass
+    __assets_supported__ = False
 
 __all__ = [
     'authenticate',
@@ -28,45 +31,84 @@ __all__ = [
     'decorate',
     'get_document',
     'init_form',
+    'mfa_authenticate',
     'redirect',
     'render_template',
     'store_assets',
     'validate'
-    ]
+]
 
 
 def authenticate(
         user_g_key='user',
         sign_in_endpoint='manage_users.sign_in',
-        sign_out_endpoint='manage_users.sign_out'
-        ):
+        sign_out_endpoint='manage_users.sign_out',
+        mfa_required=None,
+        mfa_authorize_endpoint='manage_users.mfa_auth',
+        mfa_authorized_attr='mfa_authorized',
+        mfa_enable_endpoint='manage_users.mfa_enable',
+        mfa_enabled_attr='mfa_enabled',
+    ):
     """
     Authenticate the caller has permission to call the view.
 
-    This link adds a `user` key to the the state containing the currently signed
-    in user.
+    This link adds a `user` key to the the state containing the currently
+    signed in user.
     """
 
     def authenticate(state):
         # Get the signed in user
         state.manage_user = flask.g.get(user_g_key)
 
+        # Extract the full path (we use this approach to cater for the
+        # dispatcher potentially modifying the URL root value).
+        parts = urlparse.urlparse(flask.request.url)
+        full_path = flask.request.url.split(parts[1], 1)[1]
+
         # We're not allowed to access this view point so determine if that's
         # because we're not sign-in or because we don't have permission.
         if not state.manage_user:
             # We need to sign-in to view this endpoint
 
-            # Forward the user to the sign-in page with the requested URL as the
-            # `next` parameter.
+            # Forward the user to the sign-in page with the requested URL as
+            # the `next` parameter.
             redirect_url = flask.url_for(
                 sign_in_endpoint,
                 next=secure.safe_redirect_url(
-                    flask.request.url,
+                    full_path,
                     [flask.url_for(sign_out_endpoint)]
-                    )
                 )
+            )
             return flask.redirect(redirect_url)
 
+        # Check that any user with multi-factor authentication (MFA) enabled
+        # has authorized their session using MFA.
+        if getattr(state.manage_user, mfa_enabled_attr, None):
+            if not getattr(state.manage_user, mfa_authorized_attr, None):
+
+                # The user needs to MFA authorize their session to view this
+                # endpoint.
+
+                # Forward the user to the MFA auth page with the requested URL
+                # as the `caller` parameter.
+                redirect_url = flask.url_for(
+                    mfa_authorize_endpoint,
+                    caller_url=full_path
+                )
+
+                return flask.redirect(redirect_url)
+
+        # Check if the application requires multi-factor authentication (MFA)
+        # to be enabled for users. If so and a user doesn't have MFA enabled
+        # then we redirect them to enable MFA.
+        _mfa_required = mfa_required
+        if _mfa_required is None:
+            _mfa_required = flask.current_app.config.get('USER_MFA_REQUIRED')
+
+        if _mfa_required:
+            if not getattr(state.manage_user, mfa_enabled_attr, None):
+                return flask.redirect(flask.url_for(mfa_enable_endpoint))
+
         # Check if we're allowed to access this requested endpoint
         if not Nav.allowed(flask.request.endpoint, **flask.request.view_args):
 
@@ -98,7 +140,7 @@ def config(**defaults):
 
     return config
 
-def decorate(view_type, title=None, last_breadcrumb=None, tabs=None):
+def decorate(view_type, title=None, last_breadcrumb=None):
     """
     Return a function that will add decor information to the state for the
     named view.
@@ -207,6 +249,97 @@ def init_form(populate_on=None):
 
     return init_form
 
+def mfa_authenticate_scoped_session(
+    authorize_endpoint='manage_users.mfa_auth',
+    enabled_attr='mfa_enabled',
+    get_cache=None
+):
+    """
+    Return a function that creates an authorized 2-factor authenticated
+    session.
+
+    This link should be added after the `authenticate` link.
+    """
+
+    if get_cache is None:
+        get_cache = lambda s: \
+                flask.current_app.config['USER_MFA_SCOPED_SESSION_CACHE']
+
+    def mfa_authenticate_scoped_session(state):
+
+        if getattr(state.manage_user, enabled_attr, None):
+
+            # Generate the caller URL
+
+            # Extract the full path (we use this approach to cater for the
+            # dispatcher potentially modifying the URL root value).
+            parts = urlparse.urlparse(flask.request.url)
+            full_path = flask.request.url.split(parts[1], 1)[1]
+
+            # Parse URL
+            caller_url_parts = list(urlparse.urlparse(full_path))
+
+            # Remove 'mfa_scoped_session_token' from the query parameters
+            query = dict(urlparse.parse_qsl(caller_url_parts[4]))
+            query.pop('mfa_scoped_session_token', None)
+            caller_url_parts[4] = urlencode(query)
+
+            # Rebuild the URL
+            caller_url = urlparse.urlunparse(caller_url_parts)
+
+            # Check if a scoped session has been authorized using MFA
+            session_token \
+                    = flask.request.values.get('mfa_scoped_session_token')
+
+            if session_token:
+
+                # Verify the scoped session
+                verified = secure.mfa.verify_scoped_session(
+                    get_cache(state),
+                    session_token,
+                    (str(state.manage_user._id), caller_url)
+                )
+                if verified:
+                    return
+
+                # Notify the user that their session likely expired
+                flask.flash('2FA session expired', 'error')
+
+            # Redirect the user to verify the session
+            return flask.redirect(
+                flask.url_for(
+                    authorize_endpoint,
+                    caller_url=caller_url,
+                    scoped=True
+                )
+            )
+
+    return mfa_authenticate_scoped_session
+
+def mfa_end_scoped_session(
+    authorize_endpoint='manage_users.mfa_auth',
+    enabled_attr='mfa_enabled',
+    get_cache=None
+):
+    """
+    Return a function that will clear a scoped session.
+
+    This link should be added as the penultimate link, e.g before `redirect`
+    or `render_template`.
+    """
+
+    if get_cache is None:
+        get_cache = lambda s: \
+                flask.current_app.config['USER_MFA_SCOPED_SESSION_CACHE']
+
+    def mfa_end_scoped_session(state):
+        secure.mfa.delete_scoped_session(
+            get_cache,
+            flask.request.values.get('mfa_scoped_session_token')
+        )
+
+    return mfa_end_scoped_session
+
 def redirect(endpoint, include_id=False):
     """
     Return a function that will trigger a redirect to the given endpoint.
@@ -260,117 +393,124 @@ def store_assets():
 
     def store_assets(state):
 
-        # Check that the app supports an asset manager, if not then there's
-        # nothing to do.
-        if not hasattr(flask.current_app, 'asset_mgr'):
+        if not getattr(flask.current_app, 'asset_mgr', None):
             return
-        asset_mgr = flask.current_app.asset_mgr
 
-        # Get the document being added or updated
+        # Identify asset fields within the form
         document = state[state.manage_config.var_name]
-
-        # Look for asset fields against the document, convert temporary assets
-        # to permenant assets and check for changes to the base transform which
-        # require any existing variations for the asset to be re-generated.
-
-        new_assets = []
-        updated_assets = []
-        removed_assets = []
+        asset_fields = []
 
         for field in document.get_fields():
 
-            # This must relate to a field within the form
-            if not getattr(state.form, field, None):
+            if not isinstance(getattr(state.form, field, None), AssetField):
                 continue
 
-            # We're only interested in fields that hold assets
-            if not isinstance(state.form[field], AssetField):
-                continue
+            asset_fields.append(field)
 
-            value = state.form_data.get(field)
+        # Build a list of assets to make permanent (persist), to generate new
+        # variations for (transform) and to remove.
+        assets_to_persist = []
+        assets_to_transform = []
+        assets_to_remove = []
 
-            # Check if the asset is new
-            if value:
-                if value.temporary:
+        for field in asset_fields:
 
-                    # Store the asset permenantly
-                    flask.current_app.asset_mgr.store(value)
+            asset = state.form_data.get(field)
 
-                    # Log that we need to
-                    new_assets.append(field)
+            if asset:
 
+                if asset.temporary:
+                    assets_to_persist.append(asset)
+
+                    if not asset.base_transforms:
+                        continue
+
+                elif not state.form[field].base_transform_modified:
                     continue
 
-                # Check if the base transforms for the asset have been modified
-                if state.form[field].base_transform_modified:
-                    updated_assets.append(field)
-
             else:
-                removed_assets.append(field)
 
-        # Generate variations for new and updated assets
-        for field in (new_assets + updated_assets):
-            value = state.form_data.get(field)
+                if document.get(field):
+                    assets_to_remove.append(document[field])
 
-            variations = {}
+                continue
 
-            # Check for variations defined against the manage config
-            if hasattr(state.manage_config, field + '_variations'):
-                variations = getattr(
-                    state.manage_config,
-                    field + '_variations'
-                )
+            # Build the transform instructions required to regenerate
+            # variations for the asset.
 
-            # Check for existing variations (which overide any variation set
-            # against the manage config.
-            if value.variations:
-                for name, variation_asset in value.variations.items():
-                    variations[name] = [BaseTransform.from_json_type(t)
-                        for t in variation_asset.local_transforms]
+            # Get variations defined for the gallery field against the manage
+            # config.
+            variations = state.manage_config.asset_variations.get(field, {})
 
-            # Ensure the draft variation is never updated
-            if '--draft--' in variations:
-                variations.pop('--draft--')
+            if asset.variations:
+
+                # Check for local transforms against the asset which are added
+                # ro and may override variations defined for the gallery
+                # field.
+                for variation_name, variation_asset \
+                        in asset.variations.items():
+
+                    if not isinstance(variation_asset, Asset):
+                        variation_asset = Asset(variation_asset)
+
+                    variations[variation_name] = [
+                        BaseTransform.from_json_type(t)
+                        for t in variation_asset.local_transforms
+                    ]
+
+            # Ensure the system set '--draft--' variation is never updated
+            variations.pop('--draft--', None)
 
             if variations:
 
-                # Get the list of base transforms for the asset
-                base_transforms = [BaseTransform.from_json_type(t)
-                        for t in value.base_transforms]
-
-                # Store variations for the asset
-                asset_mgr.generate_variations(
-                    value,
+                # Add the tranform information for the asset (the asset,
+                # variations and base transforms).
+                assets_to_transform.append((
+                    asset,
                     variations,
-                    base_transforms
-                )
+                    [
+                        BaseTransform.from_json_type(t)
+                        for t in asset.base_transforms
+                    ]
+                ))
 
-        # Save any changes to the database
-        update_fields = new_assets + updated_assets + removed_assets
+        # Store assets permanently
+        asset_mgr = flask.current_app.asset_mgr
+        asset_mgr.persist_many(assets_to_persist)
 
-        if update_fields:
+        # Generate variations
+        asset_mgr.generate_variations_for_many(
+            [a[0] for a in assets_to_transform],
+            {a[0].key: a[1] for a in assets_to_transform},
+            {a[0].key: a[2] for a in assets_to_transform}
+        )
 
-            # Check to see if the frame class supports `logged_update`s and if
-            # so...
-            if hasattr(state.manage_config.frame_cls, 'logged_update'):
+        if asset_fields:
 
-                # Supports `logged_update`
+            # Update the document
+            if hasattr(document, 'logged_update'):
                 document.logged_update(
                     state.manage_user,
                     {
-                        f: state.form_data.get(f)._document \
-                        if state.form_data.get(f) else None \
-                        for f in update_fields
+                        field: state.form_data.get(field) or None
+                        for field in asset_fields
                     }
                 )
 
             else:
+                for field in asset_fields:
+                    setattr(
+                        document,
+                        field,
+                        state.form_data.get(field) or None
+                    )
 
-                # Doesn't support `logged_update`
-                for field in update_fields:
-                    setattr(document, field, state.form_data.get(field))
+                document.update(*asset_fields)
 
-                document.update(*update_fields)
+        if state.remove_assets:
+
+            # Remove assets
+            asset_mgr.remove_many(assets_to_remove)
 
     return store_assets
 
